from io import StringIO
from textwrap import dedent
from unittest.mock import patch

import responses
from django.core.management import CommandError, call_command
from django.utils import timezone

from arkindex.process.models import ArkindexFeature, FeatureUsage, Worker, WorkerType, WorkerVersion, WorkerVersionState
from arkindex.project.tests import ArkindexTestCase


def mock_config():
    return {
    "features": {
        "file_import": {"image": "registry.example.com/file-import:latest", "command": None},
        "init_elements":
            {"image": "registry.example.com/init-elements:latest", "command": "init-elements"},
        "s3_ingest": {"image": "registry.example.com/s3-ingest:latest", "command": None},
        "pdf_export": {"image": "registry.example.com/file-export:latest", "command": "worker-export-pdf"},
        "pagexml_export": {"image": "registry.example.com/file-export:latest", "command": "worker-export-pagexml"}
    }
}

def mock_teklia_worker_config():
    return {
        "features": {
            "file_import": {
                "teklia_worker": {
                    "name": "repository/import/file",
                    "version": "0.1.0",
                    "slug": "file-import",
                }
            }
        }
    }


class TestUpdateSystemWorkers(ArkindexTestCase):

    def setUp(self):
        super().setUp()
        self.maxDiff = None

    def update_system_workers(self):
        output = StringIO()
        call_command(
            "update_system_workers",
            ["--no-color"],
            stdout=output,
            stderr=output,
        )
        return output.getvalue().strip()

    def check_feature_version(self, worker_version, feature):
        """
        Assert that a WorkerVersion has the expected attributes to provide a given feature
        """
        self.assertEqual(worker_version.feature, feature)
        self.assertEqual(worker_version.docker_image_iid, mock_config()["features"][feature.value]["image"])
        self.assertEqual(worker_version.docker_command, mock_config()["features"][feature.value]["command"])
        self.assertEqual(worker_version.state, WorkerVersionState.Available)
        self.assertEqual(worker_version.required_user_configuration_fields, set())
        self.assertEqual(worker_version.gpu_usage, FeatureUsage.Disabled)
        self.assertEqual(worker_version.model_usage, FeatureUsage.Disabled)

        self.assertIsNone(worker_version.worker.archived)
        self.assertTrue(worker_version.worker.public)

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_from_scratch(self):
        output = self.update_system_workers()

        worker_type = WorkerType.objects.get()
        self.assertEqual(worker_type.slug, "system")
        self.assertEqual(worker_type.display_name, "System")

        file_import_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)

        pagexml_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML)
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

        # Check the attributes on the new workers
        self.assertEqual(file_import_version.worker.name, "FileImport")
        self.assertEqual(file_import_version.worker.slug, "file_import")
        self.assertEqual(file_import_version.worker.type, worker_type)
        self.assertEqual(init_elements_version.worker.name, "InitElements")
        self.assertEqual(init_elements_version.worker.slug, "init_elements")
        self.assertEqual(init_elements_version.worker.type, worker_type)
        self.assertEqual(s3_ingest_version.worker.name, "S3Ingest")
        self.assertEqual(s3_ingest_version.worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_version.worker.type, worker_type)
        self.assertEqual(pdf_export_version.worker.name, "ExportPDF")
        self.assertEqual(pdf_export_version.worker.slug, "pdf_export")
        self.assertEqual(pdf_export_version.worker.type, worker_type)
        self.assertEqual(pagexml_export_version.worker.name, "ExportPageXML")
        self.assertEqual(pagexml_export_version.worker.slug, "pagexml_export")
        self.assertEqual(pagexml_export_version.worker.type, worker_type)

        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Created new System worker type ({worker_type.id})
            Created new FileImport system worker
            Creating new worker version
            Using new worker version {file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Created new InitElements system worker
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Created new ExportPDF system worker
            Creating new worker version
            Using new worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Created new ExportPageXML system worker
            Creating new worker version
            Using new worker version {pagexml_export_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_existing_worker_type(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)

        file_import_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)

        pagexml_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML)
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

        # Check the attributes on the two new workers
        self.assertEqual(init_elements_version.worker.name, "InitElements")
        self.assertEqual(init_elements_version.worker.slug, "init_elements")
        self.assertEqual(init_elements_version.worker.type, worker_type)
        self.assertEqual(s3_ingest_version.worker.name, "S3Ingest")
        self.assertEqual(s3_ingest_version.worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_version.worker.type, worker_type)
        self.assertEqual(pdf_export_version.worker.name, "ExportPDF")
        self.assertEqual(pdf_export_version.worker.slug, "pdf_export")
        self.assertEqual(pdf_export_version.worker.type, worker_type)
        self.assertEqual(pagexml_export_version.worker.name, "ExportPageXML")
        self.assertEqual(pagexml_export_version.worker.slug, "pagexml_export")
        self.assertEqual(pagexml_export_version.worker.type, worker_type)

        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Created new FileImport system worker
            Creating new worker version
            Using new worker version {file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Created new InitElements system worker
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Created new ExportPDF system worker
            Creating new worker version
            Using new worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Created new ExportPageXML system worker
            Creating new worker version
            Using new worker version {pagexml_export_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_existing_system_workers(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
        )
        pdf_export_worker = Worker.objects.create(
            type=worker_type,
            slug="pdf_export",
            name="PDF, export of the",
            public=True
        )
        pagexml_export_worker = Worker.objects.create(
            type=worker_type,
            slug="pagexml_export",
            name="Page XML, der Export",
            public=True
        )

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 5)

        file_import_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.assertEqual(file_import_version.worker, file_import_worker)
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.assertEqual(init_elements_version.worker, init_elements_worker)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.assertEqual(s3_ingest_version.worker, s3_ingest_worker)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        self.assertEqual(pdf_export_version.worker, pdf_export_worker)
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)

        pagexml_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML)
        self.assertEqual(pagexml_export_version.worker, pagexml_export_worker)
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Using existing system worker Sir File-a-Lot
            Worker is up to date
            Creating new worker version
            Using new worker version {file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Worker is up to date
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Worker is up to date
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Using existing system worker PDF, export of the
            Worker is up to date
            Creating new worker version
            Using new worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Using existing system worker Page XML, der Export
            Worker is up to date
            Creating new worker version
            Using new worker version {pagexml_export_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_worker_slug_conflict(self):
        worker_type = WorkerType.objects.create(slug="systemnt", display_name="Systemn't")

        worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Troublemaker",
        )

        output = StringIO()
        with self.assertRaisesMessage(CommandError, "Cannot create system worker with slug file_import due to a conflict."):
            call_command(
                "update_system_workers",
                ["--no-color"],
                stdout=output,
                stderr=output,
            )

        self.assertEqual(output.getvalue().strip(), dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Worker Troublemaker ({worker.id}) uses the same slug with a different worker type.
            Please update this worker's type to `system` or change its slug.
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_update_existing_system_workers(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            # This worker will be made public
            public=False,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
            # This worker will be unarchived
            archived=timezone.now(),
        )

        output = self.update_system_workers()

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 5)

        file_import_worker.refresh_from_db()
        self.assertEqual(file_import_worker.name, "Sir File-a-Lot")
        self.assertEqual(file_import_worker.slug, "file_import")
        self.assertEqual(file_import_worker.type, worker_type)
        self.assertTrue(file_import_worker.public)
        self.assertIsNone(file_import_worker.archived)

        init_elements_worker.refresh_from_db()
        self.assertEqual(init_elements_worker.name, "Init of the Elements")
        self.assertEqual(init_elements_worker.slug, "init_elements")
        self.assertEqual(init_elements_worker.type, worker_type)
        self.assertTrue(init_elements_worker.public)
        self.assertIsNone(init_elements_worker.archived)

        s3_ingest_worker.refresh_from_db()
        self.assertEqual(s3_ingest_worker.name, "S3, Ingest of the")
        self.assertEqual(s3_ingest_worker.slug, "s3_ingest")
        self.assertEqual(s3_ingest_worker.type, worker_type)
        self.assertTrue(s3_ingest_worker.public)
        self.assertIsNone(s3_ingest_worker.archived)

        file_import_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.assertEqual(file_import_version.worker, file_import_worker)
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)

        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        self.assertEqual(init_elements_version.worker, init_elements_worker)
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)

        s3_ingest_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest)
        self.assertEqual(s3_ingest_version.worker, s3_ingest_worker)
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)

        pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)

        pagexml_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML)
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Using existing system worker Sir File-a-Lot
            Worker is up to date
            Creating new worker version
            Using new worker version {file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Marking worker as public
            Creating new worker version
            Using new worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Unarchiving worker
            Creating new worker version
            Using new worker version {s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Created new ExportPDF system worker
            Creating new worker version
            Using new worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Created new ExportPageXML system worker
            Creating new worker version
            Using new worker version {pagexml_export_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_assigns_existing_compatible_versions(self):
        """
        The command should assign the feature to an existing WorkerVersion compatible with it,
        marking the version as available if necessary
        """
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_ingest",
            name="S3, Ingest of the",
            public=True,
        )
        pdf_export_worker = Worker.objects.create(
            type=worker_type,
            slug="pdf_export",
            name="PDF, export of the",
            public=True
        )
        pagexml_export_worker = Worker.objects.create(
            type=worker_type,
            slug="pagexml_export",
            name="Page XML, der Export",
            public=True
        )

        # Those versions use the correct image, but do not have acceptable attributes
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=1,
            configuration={"docker": {"command": "exit"}},
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=2,
            configuration={
                "user_configuration": {
                    "nobody_expects_the": {
                        "title": "required configuration",
                        "type": "int",
                        "required": True,
                    }
                },
            },
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=3,
            gpu_usage=FeatureUsage.Required,
        )
        init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=4,
            model_usage=FeatureUsage.Required,
        )
        pdf_export_worker.versions.create(
            docker_image_iid="registry.example.com/file-export:latest",
            version=2,
            configuration={"docker": {"command": "something"}}
        )

        # These versions should be assigned to the features
        file_import_version = file_import_worker.versions.create(
            docker_image_iid="registry.example.com/file-import:latest",
            version=1,
            state=WorkerVersionState.Available,
        )
        init_elements_version = init_elements_worker.versions.create(
            docker_image_iid="registry.example.com/init-elements:latest",
            version=5,
            state=WorkerVersionState.Available,
            configuration={"docker": {"command": "init-elements"}},
        )
        s3_ingest_version = s3_ingest_worker.versions.create(
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=1,
            state=WorkerVersionState.Created,
        )
        pdf_export_version = pdf_export_worker.versions.create(
            docker_image_iid="registry.example.com/file-export:latest",
            version=1,
            configuration={"docker": {"command": "worker-export-pdf"}}
        )
        pagexml_export_version = pagexml_export_worker.versions.create(
            docker_image_iid="registry.example.com/file-export:latest",
            version=1,
            configuration={"docker": {"command": "worker-export-pagexml"}}
        )

        self.assertEqual(self.update_system_workers(), dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Using existing system worker Sir File-a-Lot
            Worker is up to date
            Assigning existing worker version {file_import_version.id} to the feature
            Using existing worker version {file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Using existing system worker Init of the Elements
            Worker is up to date
            Assigning existing worker version {init_elements_version.id} to the feature
            Setup docker command: init-elements
            Using existing worker version {init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Using existing system worker S3, Ingest of the
            Worker is up to date
            Assigning existing worker version {s3_ingest_version.id} to the feature
            Marking the worker version as available
            Using existing worker version {s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Using existing system worker PDF, export of the
            Worker is up to date
            Assigning existing worker version {pdf_export_version.id} to the feature
            Setup docker command: worker-export-pdf
            Marking the worker version as available
            Using existing worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Using existing system worker Page XML, der Export
            Worker is up to date
            Assigning existing worker version {pagexml_export_version.id} to the feature
            Setup docker command: worker-export-pagexml
            Marking the worker version as available
            Using existing worker version {pagexml_export_version.id}
            """
        ).strip())

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 5)
        self.assertEqual(WorkerVersion.objects.count(), 10)

        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport), file_import_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements), init_elements_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest), s3_ingest_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF), pdf_export_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML), pagexml_export_version)

        file_import_version.refresh_from_db()
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)
        init_elements_version.refresh_from_db()
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)
        s3_ingest_version.refresh_from_db()
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)
        pdf_export_version.refresh_from_db()
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)
        pagexml_export_version.refresh_from_db()
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_unassigns_incompatible_versions(self):
        worker_type = WorkerType.objects.create(slug="systemnt", display_name="Systemn't")
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        init_elements_worker = Worker.objects.create(
            type=worker_type,
            slug="init_elements",
            name="Init of the Elements",
            public=True,
        )
        s3_ingest_worker = Worker.objects.create(
            type=worker_type,
            slug="s3_digest",
            name="S3Digest",
            public=True,
            # This isn't a system worker, so this worker being archived means the command
            # should create a new system worker instead of reusing this one
            archived=timezone.now(),
        )
        self.assertFalse(Worker.objects.filter(slug="s3_ingest").exists())

        file_import_version = file_import_worker.versions.create(
            feature=ArkindexFeature.FileImport,
            docker_image_iid="registry.example.com/file-export:latest",
            version=1,
            state=WorkerVersionState.Available,
        )
        init_elements_version = init_elements_worker.versions.create(
            version=1,
            feature=ArkindexFeature.InitElements,
            docker_image_iid="registry.gitlab.teklia.com/callico/callico:latest",
            state=WorkerVersionState.Available,
            configuration={
                "docker": {
                    "command": "false",
                },
                "user_configuration": {
                    "nobody_expects_the": {
                        "title": "required configuration",
                        "type": "int",
                        "required": True,
                    },
                },
            },
        )
        # Everything is valid for this version, but it is on an archived worker
        s3_ingest_version = s3_ingest_worker.versions.create(
            feature=ArkindexFeature.S3Ingest,
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=1,
            state=WorkerVersionState.Available,
        )

        output = self.update_system_workers()

        system_worker_type = WorkerType.objects.exclude(slug=worker_type.slug).get()
        self.assertEqual(system_worker_type.slug, "system")
        self.assertEqual(system_worker_type.display_name, "System")

        self.assertEqual(Worker.objects.count(), 6)
        new_s3_ingest_worker = Worker.objects.get(slug="s3_ingest")
        self.assertEqual(new_s3_ingest_worker.type, system_worker_type)
        self.assertEqual(new_s3_ingest_worker.name, "S3Ingest")

        self.assertEqual(WorkerVersion.objects.count(), 8)
        new_file_import_version = file_import_worker.versions.get(version=2)
        self.check_feature_version(new_file_import_version, ArkindexFeature.FileImport)
        new_init_elements_version = init_elements_worker.versions.get(version=2)
        self.check_feature_version(new_init_elements_version, ArkindexFeature.InitElements)
        new_s3_ingest_version = new_s3_ingest_worker.versions.get()
        self.check_feature_version(new_s3_ingest_version, ArkindexFeature.S3Ingest)

        pdf_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF)
        pagexml_export_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML)

        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Current worker version: {file_import_version.id} (registry.example.com/file-export:latest)
            This version has an invalid Docker image.
            Unassigning feature from the current version
            Worker is up to date
            Creating new worker version
            Using new worker version {new_file_import_version.id}
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Current worker version: {init_elements_version.id} (registry.gitlab.teklia.com/callico/callico:latest)
            This version has an invalid Docker image.
            This version uses a custom Docker command which could interfere with the feature.
            This version uses a custom configuration which could interfere with the feature.
            Unassigning feature from the current version
            Worker is up to date
            Creating new worker version
            Using new worker version {new_init_elements_version.id}
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Current worker version: {s3_ingest_version.id} (registry.example.com/s3-ingest:latest)
            This version is part of an archived worker.
            Unassigning feature from the current version
            Created new System worker type ({system_worker_type.id})
            Created new S3Ingest system worker
            Creating new worker version
            Using new worker version {new_s3_ingest_version.id}
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Created new ExportPDF system worker
            Creating new worker version
            Using new worker version {pdf_export_version.id}
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Created new ExportPageXML system worker
            Creating new worker version
            Using new worker version {pagexml_export_version.id}
            """
        ).strip())

    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_config)
    def test_noop(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        worker = Worker.objects.create(type=worker_type, slug="worker", name="Worker", public=True)

        file_import_version = worker.versions.create(
            feature=ArkindexFeature.FileImport,
            docker_image_iid="registry.example.com/file-import:latest",
            version=1,
            state=WorkerVersionState.Available,
        )
        init_elements_version = worker.versions.create(
            feature=ArkindexFeature.InitElements,
            docker_image_iid="registry.example.com/init-elements:latest",
            configuration={"docker": {"command": "init-elements"}},
            version=2,
            state=WorkerVersionState.Available,
        )
        s3_ingest_version = worker.versions.create(
            feature=ArkindexFeature.S3Ingest,
            docker_image_iid="registry.example.com/s3-ingest:latest",
            version=3,
            state=WorkerVersionState.Available,
        )
        pdf_export_version = worker.versions.create(
            feature=ArkindexFeature.ExportPDF,
            docker_image_iid="registry.example.com/file-export:latest",
            configuration={"docker": {"command": "worker-export-pdf"}},
            version=4,
            state=WorkerVersionState.Available
        )
        pagexml_export_version = worker.versions.create(
            feature=ArkindexFeature.ExportPageXML,
            docker_image_iid="registry.example.com/file-export:latest",
            configuration={"docker": {"command": "worker-export-pagexml"}},
            version=5,
            state=WorkerVersionState.Available
        )

        self.assertEqual(self.update_system_workers(), dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Using registry.example.com/file-import:latest to provide FileImport
            Current worker version: {file_import_version.id} (registry.example.com/file-import:latest)
            Worker is up to date
            Worker version for FileImport is up to date
            ───────────────────────────────── InitElements ─────────────────────────────────
            Using registry.example.com/init-elements:latest to provide InitElements
            Current worker version: {init_elements_version.id} (registry.example.com/init-elements:latest)
            This version uses the required Docker command.
            Worker is up to date
            Worker version for InitElements is up to date
            ─────────────────────────────────── S3Ingest ───────────────────────────────────
            Using registry.example.com/s3-ingest:latest to provide S3Ingest
            Current worker version: {s3_ingest_version.id} (registry.example.com/s3-ingest:latest)
            Worker is up to date
            Worker version for S3Ingest is up to date
            ────────────────────────────────── ExportPDF ───────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPDF
            Current worker version: {pdf_export_version.id} (registry.example.com/file-export:latest)
            This version uses the required Docker command.
            Worker is up to date
            Worker version for ExportPDF is up to date
            ──────────────────────────────── ExportPageXML ─────────────────────────────────
            Using registry.example.com/file-export:latest to provide ExportPageXML
            Current worker version: {pagexml_export_version.id} (registry.example.com/file-export:latest)
            This version uses the required Docker command.
            Worker is up to date
            Worker version for ExportPageXML is up to date
            """
        ).strip())

        self.assertEqual(WorkerType.objects.count(), 1)
        self.assertEqual(Worker.objects.count(), 1)
        self.assertEqual(WorkerVersion.objects.count(), 5)

        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport), file_import_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements), init_elements_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.S3Ingest), s3_ingest_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPDF), pdf_export_version)
        self.assertEqual(WorkerVersion.objects.get_by_feature(ArkindexFeature.ExportPageXML), pagexml_export_version)

        file_import_version.refresh_from_db()
        self.check_feature_version(file_import_version, ArkindexFeature.FileImport)
        init_elements_version.refresh_from_db()
        self.check_feature_version(init_elements_version, ArkindexFeature.InitElements)
        s3_ingest_version.refresh_from_db()
        self.check_feature_version(s3_ingest_version, ArkindexFeature.S3Ingest)
        pdf_export_version.refresh_from_db()
        self.check_feature_version(pdf_export_version, ArkindexFeature.ExportPDF)
        pagexml_export_version.refresh_from_db()
        self.check_feature_version(pagexml_export_version, ArkindexFeature.ExportPageXML)

    @responses.activate
    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_teklia_worker_config)
    def test_teklia_worker_invalid_slug(self):
        responses.add(
            responses.GET,
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml",
            json={"version": 2, "workers": [{"slug": "no"}]},
          )
        with self.assertRaisesMessage(CommandError, (
            "No worker with slug file-import in .arkindex.yml at "
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml."
        )):
            self.update_system_workers()

    @responses.activate
    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_teklia_worker_config)
    def test_teklia_worker_invalid_yaml(self):
        responses.add(
            responses.GET,
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml",
            json={"version": 42}
          )
        with self.assertRaisesMessage(CommandError, (
            "Error retrieving configuration at "
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml: "
            "invalid YAML configuration"
        )):
            self.update_system_workers()

    @responses.activate
    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_teklia_worker_config)
    def test_teklia_worker_from_scratch(self):
        responses.add(
            responses.GET,
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml",
            json={
                "version": 2,
                "workers": [{
                    "slug": "file-import",
                    "name": "File Import",
                    "type": "import_type",
                    "description": "test",
                }],
            },
          )

        with self.assertRaises(WorkerVersion.DoesNotExist):
            WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)

        output = self.update_system_workers()
        worker_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Configuring feature Fileimport from https://gitlab.teklia.com/repository/import/file
            Using registry.gitlab.teklia.com/repository/import/file/0.1.0 to provide FileImport
            Created new System worker type ({worker_version.worker.type_id})
            Created new FileImport system worker
            Creating new worker version
            Using new worker version {worker_version.id}
            """
        ).strip())
        self.assertEqual(worker_version.configuration, {
            "description": "test",
            "name": "File Import",
            "slug": "file-import",
            "type": "import_type",
        })
        self.assertEqual(worker_version.docker_image_iid, "registry.gitlab.teklia.com/repository/import/file/0.1.0")
        self.assertEqual(worker_version.feature, ArkindexFeature.FileImport)
        self.assertEqual(worker_version.state, WorkerVersionState.Available)
        self.assertEqual(worker_version.version, 1)
        self.assertEqual(worker_version.docker_image_iid, "registry.gitlab.teklia.com/repository/import/file/0.1.0")
        self.assertEqual(worker_version.worker.name, "FileImport")
        self.assertEqual(worker_version.worker.repository_url, "https://gitlab.teklia.com/repository/import/file")

        # Retrying does not averride the version
        self.assertEqual(self.update_system_workers(), dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Configuring feature Fileimport from https://gitlab.teklia.com/repository/import/file
            Using registry.gitlab.teklia.com/repository/import/file/0.1.0 to provide FileImport
            Current worker version: {worker_version.id} (registry.gitlab.teklia.com/repository/import/file/0.1.0)
            Worker is up to date
            Worker version for FileImport is up to date
            """
        ).strip())

    @responses.activate
    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_teklia_worker_config)
    def test_teklia_worker_unassign(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        version = file_import_worker.versions.create(
            version=1,
            docker_image_iid="test",
            state=WorkerVersionState.Available,
            feature=ArkindexFeature.FileImport,
        )
        responses.add(
            responses.GET,
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml",
            json={
                "version": 2,
                "workers": [{
                    "slug": "file-import",
                    "name": "File Import",
                    "type": "import_type",
                    "description": "test",
                }],
            },
          )

        output = self.update_system_workers()
        worker_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Configuring feature Fileimport from https://gitlab.teklia.com/repository/import/file
            Using registry.gitlab.teklia.com/repository/import/file/0.1.0 to provide FileImport
            Current worker version: {version.id} (test)
            This version has an invalid Docker image.
            This version uses a custom configuration which could interfere with the feature.
            Unassigning feature from the current version
            Worker is up to date
            Creating new worker version
            Using new worker version {worker_version.id}
            """
        ).strip())
        self.assertEqual(worker_version.configuration, {
            "description": "test",
            "name": "File Import",
            "slug": "file-import",
            "type": "import_type",
        })
        self.assertEqual(worker_version.docker_image_iid, "registry.gitlab.teklia.com/repository/import/file/0.1.0")
        self.assertEqual(worker_version.feature, ArkindexFeature.FileImport)
        self.assertEqual(worker_version.state, WorkerVersionState.Available)
        self.assertEqual(worker_version.version, 2)
        self.assertEqual(worker_version.worker.name, "Sir File-a-Lot")
        self.assertEqual(worker_version.worker.repository_url, None)

    @responses.activate
    @patch("arkindex.process.management.commands.update_system_workers.parse_config", mock_teklia_worker_config)
    def test_teklia_worker_existing(self):
        worker_type = WorkerType.objects.create(slug="system", display_name="Système")
        worker_conf = {
            "slug": "file-import",
            "name": "File Import",
            "type": "import_type",
            "description": "test",
        }
        file_import_worker = Worker.objects.create(
            type=worker_type,
            slug="file_import",
            name="Sir File-a-Lot",
            public=True,
        )
        version = file_import_worker.versions.create(
            version=1,
            docker_image_iid="registry.gitlab.teklia.com/repository/import/file/0.1.0",
            state=WorkerVersionState.Available,
            feature=ArkindexFeature.FileImport,
            configuration=worker_conf,
        )
        responses.add(
            responses.GET,
            "https://gitlab.teklia.com/repository/import/file/-/raw/0.1.0/.arkindex.yml",
            json={"version": 2, "workers": [worker_conf]},
          )

        output = self.update_system_workers()
        self.assertEqual(WorkerVersion.objects.count(), 1)
        worker_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.FileImport)
        self.assertEqual(output, dedent(
            f"""
            ────────────────────────────────── FileImport ──────────────────────────────────
            Configuring feature Fileimport from https://gitlab.teklia.com/repository/import/file
            Using registry.gitlab.teklia.com/repository/import/file/0.1.0 to provide FileImport
            Current worker version: {version.id} (registry.gitlab.teklia.com/repository/import/file/0.1.0)
            Worker is up to date
            Worker version for FileImport is up to date
            """
        ).strip())
        self.assertEqual(worker_version.docker_image_iid, "registry.gitlab.teklia.com/repository/import/file/0.1.0")
        self.assertEqual(worker_version.feature, ArkindexFeature.FileImport)
        self.assertEqual(worker_version.state, WorkerVersionState.Available)
        self.assertEqual(worker_version.version, 1)
        self.assertEqual(worker_version.worker.name, "Sir File-a-Lot")
        self.assertEqual(worker_version.worker.repository_url, None)
