
from arkindex.process.builder import ProcessBuilder
from arkindex.process.models import (
    ArkindexFeature,
    ProcessMode,
    WorkerConfiguration,
    WorkerVersion,
)
from arkindex.project.tests import FixtureAPITestCase


class TestProcessName(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")
        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.version_gpu = WorkerVersion.objects.get(worker__slug="worker-gpu")
        cls.version_with_model = WorkerVersion.objects.get(worker__slug="generic")

        cls.reco_config_1 = WorkerConfiguration.objects.create(
            worker=cls.recognizer.worker,
            name="Config 1",
            configuration={"param1": "value1"},
        )
        cls.reco_config_2 = WorkerConfiguration.objects.create(
            worker=cls.recognizer.worker,
            name="Config 2",
            configuration={"param1": "value2"},
        )
        cls.reco_config_3 = WorkerConfiguration.objects.create(
            worker=cls.recognizer.worker,
            name="Config 3",
            configuration={"param1": "value3"},
        )
        cls.dla_config = WorkerConfiguration.objects.create(
            worker=cls.dla.worker,
            name="Tellus stabilita",
            configuration={"param": "value"},
        )

        cls.workers_process = cls.corpus.processes.create(creator=cls.user, mode=ProcessMode.Workers)

    def test_default_process_name(self):
        """
        If a workers mode process gets started with no process name set, a default process name
        is built from its worker runs. The elements initialisation worker run is ignored.
        """
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.workers_process.worker_runs.create(
            version=init_elements_version,
            ttl=0,
        )
        self.workers_process.worker_runs.create(
            version=self.recognizer,
            parents=[init_run.id],
            ttl=0,
        )
        dla_run = self.workers_process.worker_runs.create(
            version=self.dla,
            parents=[init_run.id],
            ttl=0,
        )
        self.workers_process.worker_runs.create(
            version=self.version_gpu,
            parents=[dla_run.id],
            ttl=0,
        )
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, None)

        builder = ProcessBuilder(process=self.workers_process)
        builder.build()
        builder.save()
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, "Document layout analyser, Recognizer, Worker requiring a GPU")

    def test_default_process_name_same_worker(self):
        """
        If the process has multiple worker runs using the same worker:
        - if they are at the same depth, the worker name is not added again to the process name
        - if they are at different depths, the worker name is added to the process name

             reco - reco - reco
            /
        init
            \
             dla - reco - dla
        """
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.workers_process.worker_runs.create(
            version=init_elements_version,
            ttl=0,
        )
        reco_run_1 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            parents=[init_run.id],
            ttl=0,
        )
        reco_run_2 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.reco_config_1,
            parents=[reco_run_1.id],
            ttl=0,
        )
        self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.reco_config_2,
            parents=[reco_run_2.id],
            ttl=0,
        )
        dla_run_1 = self.workers_process.worker_runs.create(
            version=self.dla,
            parents=[init_run.id],
            ttl=0,
        )
        reco_run_4 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.reco_config_3,
            parents=[dla_run_1.id],
            ttl=0,
        )
        self.workers_process.worker_runs.create(
            version=self.dla,
            configuration=self.dla_config,
            parents=[reco_run_4.id],
            ttl=0,
        )

        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, None)

        builder = ProcessBuilder(process=self.workers_process)
        builder.build()
        builder.save()
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, "Document layout analyser, Recognizer, Recognizer, Document layout analyser, Recognizer")

    def test_default_process_name_truncated(self):
        """
        If the default process worker name is too long (len() > 250) it gets truncated
        """
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.workers_process.worker_runs.create(version=init_elements_version, ttl=0)
        # Update the recognizer worker's name so that it is long
        self.recognizer.worker.name = "animula vagula blandula hospes comesque corporis quae nunc abibis in loca pallidula rigida nudula ne"
        self.recognizer.worker.save()
        reco_run_1 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            parents=[init_run.id],
            ttl=0,
        )
        reco_run_2 = self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.reco_config_1,
            parents=[reco_run_1.id],
            ttl=0,
        )
        self.workers_process.worker_runs.create(
            version=self.recognizer,
            configuration=self.reco_config_2,
            parents=[reco_run_2.id],
            ttl=0,
        )
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, None)

        builder = ProcessBuilder(process=self.workers_process)
        builder.build()
        builder.save()
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, "animula vagula blandula hospes comesque corporis quae nunc abibis in loca pallidula rigida nudula ne, animula vagula blandula hospes comesque corporis quae nunc abibis in loca pallidula rigida nudula ne, animula vagula blandula hospes comesque corpo…")

    def test_process_name_alread_set(self):
        """
        If a process already has a name, it doesn't get updated
        """
        self.workers_process.name = "My process"
        self.workers_process.save()
        init_elements_version = WorkerVersion.objects.get_by_feature(ArkindexFeature.InitElements)
        init_run = self.workers_process.worker_runs.create(version=init_elements_version, ttl=0)
        self.workers_process.worker_runs.create(version=self.recognizer, parents=[init_run.id], ttl=0)
        self.assertEqual(self.workers_process.name, "My process")

        builder = ProcessBuilder(process=self.workers_process)
        builder.build()
        builder.save()
        self.workers_process.save()

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.name, "My process")
